package core

import (
	"math"

	"git.urbach.dev/cli/q/src/asm"
	"git.urbach.dev/cli/q/src/cpu"
	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/expression"
	"git.urbach.dev/cli/q/src/token"
	"git.urbach.dev/cli/q/src/types"
)

// ArrayElementToRegister moves the value of an array element into the given register.
func (f *Function) ArrayElementToRegister(node *expression.Expression, register cpu.Register) (types.Type, error) {
	name := node.Children[0].Token.Text(f.File.Bytes)
	array := f.VariableByName(name)

	if array == nil {
		return nil, errors.New(&errors.UnknownIdentifier{Name: name}, f.File, node.Children[0].Token.Position)
	}

	defer f.UseVariable(array)
	index := node.Children[1]

	memory := asm.Memory{
		Base:           array.Register,
		Offset:         0,
		OffsetRegister: math.MaxUint8,
		Length:         byte(1),
	}

	switch {
	case index.Token.IsNumeric():
		offset, err := f.ToNumber(index.Token)

		if err != nil {
			return nil, err
		}

		memory.Offset = int8(offset)

	case index.Token.Kind == token.Identifier:
		indexName := index.Token.Text(f.File.Bytes)
		indexVariable := f.VariableByName(indexName)

		if indexVariable == nil {
			return nil, errors.New(&errors.UnknownIdentifier{Name: indexName}, f.File, index.Token.Position)
		}

		defer f.UseVariable(indexVariable)

		if !types.Is(indexVariable.Type, types.AnyInt) {
			return nil, errors.New(&errors.TypeMismatch{Encountered: indexVariable.Type.Name(), Expected: types.AnyInt.Name()}, f.File, index.Token.Position)
		}

		memory.OffsetRegister = indexVariable.Register

	default:
		typ, err := f.ExpressionToRegister(index, register)

		if err != nil {
			return nil, err
		}

		if !types.Is(typ, types.AnyInt) {
			return nil, errors.New(&errors.TypeMismatch{Encountered: typ.Name(), Expected: types.AnyInt.Name()}, f.File, index.Token.Position)
		}

		memory.OffsetRegister = register
	}

	f.MemoryRegister(asm.LOAD, memory, register)
	return types.Int, nil
}
