package core

import (
	"encoding/binary"

	"git.urbach.dev/cli/q/src/asm"
	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/eval"
	"git.urbach.dev/cli/q/src/expression"
	"git.urbach.dev/cli/q/src/token"
	"git.urbach.dev/cli/q/src/types"
)

// EvaluateLeaf evaluates a leaf expression.
func (f *Function) EvaluateLeaf(expr *expression.Expression) (eval.Value, error) {
	switch expr.Token.Kind {
	case token.Identifier:
		name := expr.Token.Text(f.File.Bytes)

		if name == "true" {
			value := eval.Value{
				Kind:   eval.Number,
				Type:   types.Bool,
				Number: 1,
			}

			return value, nil
		}

		if name == "false" {
			value := eval.Value{
				Kind:   eval.Number,
				Type:   types.Bool,
				Number: 0,
			}

			return value, nil
		}

		variable, function := f.Identifier(name)

		if variable != nil {
			f.UseVariable(variable)

			if variable.Alive == 0 {
				return variable.Value, nil
			}

			tmp := f.NewRegister()
			f.RegisterRegister(asm.MOVE, tmp, variable.Register)

			value := eval.Value{
				Kind:     eval.Register,
				Type:     variable.Type,
				Register: tmp,
			}

			return value, nil
		}

		if function != nil {
			value := eval.Value{
				Kind:  eval.Label,
				Type:  types.AnyPointer,
				Label: function.UniqueName,
			}

			return value, nil
		}

		return eval.Value{}, errors.New(&errors.UnknownIdentifier{Name: name}, f.File, expr.Token.Position)

	case token.Number, token.Rune:
		number, err := f.ToNumber(expr.Token)

		if err != nil {
			return eval.Value{}, err
		}

		value := eval.Value{
			Kind:   eval.Number,
			Type:   types.AnyInt,
			Number: number,
		}

		return value, nil

	case token.String:
		data := expr.Token.Bytes(f.File.Bytes)
		data = String(data)

		slice := make([]byte, len(data)+8+1)
		binary.LittleEndian.PutUint64(slice, uint64(len(data)))
		copy(slice[8:], data)
		label := f.AddBytes(slice)

		value := eval.Value{
			Kind:  eval.Label,
			Type:  types.String,
			Label: label,
		}

		return value, nil
	}

	return eval.Value{}, errors.New(errors.InvalidExpression, f.File, expr.Token.Position)
}
